

// Code written by Hans de Nivelle and Piotr Witkowski


#include "provergoal.h"



void provergoal::print_tptp( std::ostream& stream )
const throw( notfirstorder ) 
{
   tptp_namemap map;

   stream << "%------------------------------------------------------\n";
   stream << "%   This TPTP problem was generated by checker2008c.   \n";
   stream << "%------------------------------------------------------\n";
   stream << "\n\n";

   unsigned int nr = 0;  
   for( std::list< formula > :: const_iterator
           p = axioms. begin( );
           p != axioms. end( );
           ++ p )
   {
      stream << "fof( ";
      stream << "axiom" << nr << ", ";
      stream << "axiom, ( ";
      print_tptp( stream, *p, map );
      stream << " )).\n\n";

      ++ nr; 
   }

   stream << "\n\n";

   nr = 0;
   for( std::list< formula > :: const_iterator
           p = hypotheses. begin( );
           p != hypotheses. end( );
           ++ p )
   {
      stream << "fof( ";
      stream << "hypothesis" << nr << ", ";
      stream << "hypothesis, ( "; 
      print_tptp( stream, *p, map );
      stream << " )).\n\n";

      ++ nr; 
   }

   stream << "fof( ";
   stream << "goal, ";
   stream << "conjecture, ( ";
   print_tptp( stream, conjecture, map );
   stream << " )).\n\n";

}




void provergoal::print_tptp( std::ostream& stream,
                            formula f, tptp_namemap& map )
throw( notfirstorder ) 
{
   switch( f. getoperator( ))
   { 
   case fol_false:
      stream << "$false";
      return;
  
   case fol_true:
      stream << "$true"; 
      return;

   case fol_not:
      stream << "( ~ ";
      print_tptp( stream, * ( f. subformulas_begin( )), map );
      stream << " )"; 
      return;

   case fol_and:
   case fol_or:
      stream << "( ";
      for( std::list< formula > :: const_iterator 
              p = f. subformulas_begin( );
              p != f. subformulas_end( );
              ++ p )
      {
         if( p != f. subformulas_begin( ))
         {
            if( f. getoperator( ) == fol_and )
               stream << " & "; 
            else
               stream << " | ";
         }
         print_tptp( stream, *p, map );
      }
      stream << " )";
      return;

   case fol_implies:
   case fol_equiv: 
      stream << "( "; 
      {
         std::list< formula > :: const_iterator p = f. subformulas_begin( );
         print_tptp( stream, * ( p ++ ), map ); 
         if( f. getoperator( ) == fol_implies )
            stream << " => ";
         else
            stream << " <=> "; 
         print_tptp( stream, * ( p ++ ), map ); 
      }
      stream << " )"; 
      return; 
	  
   case fol_forall:
   case fol_exists:
      {
         unsigned int s = map. nrlocaldeclarations( ); 

         if( f. getoperator( ) == fol_forall )
            stream << "( ! [ ";
         else
            stream << "( ? [ ";
      
         for( std::list< declaration > :: const_iterator
                 v = f. variables_begin( ); 
                 v != f. variables_end( );
                 ++ v )
         {
            if( ! ( v -> isfunction ) || ( v -> arity ) != 0 )
               throw notfirstorder::notfirstorder( f, *v );

            if( v != f. variables_begin( ))
               stream << ", ";

            stream << map. declarelocal( *v );
         }

         stream << " ] : "; 
         print_tptp( stream, * ( f. subformulas_begin( )), map ); 
         stream << " )";

         map. backtrack(s); 
      }
      return;

   case fol_equals:
      {
         std::list< term > :: const_iterator p = f. subterms_begin( );
         print_tptp( stream, * ( p ++ ), map ); 
         stream << " = ";
         print_tptp( stream, * ( p ++ ), map ); 
      }
      return;

   case fol_atom:
      {
         // We have to determine the arity first, so that we can map
         // f. getpredicat( ) into the right TPTP identifier.

         unsigned int arity = 0;
         for( std::list< term > :: const_iterator
                 p = f. subterms_begin( );
                 p != f. subterms_end( );
                 ++ p )
         {
            ++ arity;
         }

         stream << map. lookup(
            declaration::make_preddecl( f. getpredicate( ), arity )); 

         if( arity )
         { 
            stream << "( "; 
            for( std::list< term > :: const_iterator
                    p = f. subterms_begin( );
                    p != f. subterms_end( );
                    ++ p )
            {
               if( p != f. subterms_begin( ))
                  stream << ", ";
               print_tptp( stream, *p, map );
            }
            stream << " )"; 
         }
      }
      return; 
   }

   ASSERT( false ); exit(0);   
}





void provergoal::print_tptp( std::ostream& stream,
                             term t, tptp_namemap& map ) throw( notfirstorder ) 
{
   
   const std::string& func = map. lookup( 
      declaration::make_funcdecl( t. getfunction( ), t. getnrsubterms( )));

   stream << func;

   if( t. getnrsubterms( ) != 0 )
   {
      stream << "( ";
      for( std::list< term > :: const_iterator
              p = t. subterms_begin( );
              p != t. subterms_end( );
              ++ p )
      {
         if( p != t. subterms_begin( ))
            stream << ", "; 
         print_tptp( stream, *p, map );
      }
      stream << " )";
   }
}



std::ostream& operator << ( std::ostream& stream, const notfirstorder& nfo )
{
   stream << "Formula " << nfo. f;
   stream << " contains non-first order quantification " << nfo. decl;
   stream << "\n";
   return stream;
}


std::ostream& operator << ( std::ostream& stream, const provergoal& pg )
{
   stream << "provergoal:\n";
   stream << "axioms:\n";

   for( std::list< formula > :: const_iterator
           p = pg. axioms. begin( );
           p != pg. axioms. end( );
           ++ p )
   {
      stream << "   " << *p << "\n";
   }
   stream << "\n\n";

   stream << "hypotheses:\n"; 
   for( std::list< formula > :: const_iterator
           p = pg. hypotheses. begin( );
           p != pg. hypotheses. end( );
           ++ p )
   {
      stream << "   " << *p << "\n";
   }
   stream << "\n\n";

   stream << "conjecture:\n"; 
   stream << "    |- ";
   stream << pg. conjecture << "\n";
   return stream;
}


